@@ -1,6 +1,8 @@
+/*	$NetBSD: arch.c,v 1.26 1997/10/03 17:58:00 thorpej Exp $	*/
+
 /*
- * Copyright (c) 1988, 1989, 1990 The Regents of the University of California.
- * Copyright (c) 1988, 1989 by Adam de Boor
+ * Copyright (c) 1988, 1989, 1990, 1993
+ *	The Regents of the University of California.  All rights reserved.
  * Copyright (c) 1989 by Berkeley Softworks
  * All rights reserved.
  *
@@ -36,9 +38,18 @@
  * SUCH DAMAGE.
  */
 
+#ifdef MAKE_BOOTSTRAP
+static char rcsid[] = "$NetBSD: arch.c,v 1.26 1997/10/03 17:58:00 thorpej Exp $";
+#else
+#include <sys/cdefs.h>
 #ifndef lint
-static char sccsid[] = "@(#)arch.c	5.7 (Berkeley) 12/28/90";
+#if 0
+static char sccsid[] = "@(#)arch.c	8.2 (Berkeley) 1/2/94";
+#else
+__RCSID("$NetBSD: arch.c,v 1.26 1997/10/03 17:58:00 thorpej Exp $");
+#endif
 #endif /* not lint */
+#endif
 
 /*-
  * arch.c --
@@ -84,23 +95,24 @@ static char sccsid[] = "@(#)arch.c	5.7 (Berkeley) 12/28/90";
  *	    	  	    	is out-of-date.
  *
  *	Arch_Init 	    	Initialize this module.
+ *
+ *	Arch_End 	    	Cleanup this module.
  */
 
 #include    <sys/types.h>
 #include    <sys/stat.h>
 #include    <sys/time.h>
+#include    <sys/param.h>
 #include    <ctype.h>
 #include    <ar.h>
-#ifndef linux
-#include <ranlib.h>
-#endif
+#include    <utime.h>
 #include    <stdio.h>
+#include    <stdlib.h>
+#include    <fcntl.h>
 #include    "make.h"
 #include    "hash.h"
-
-#ifndef RANLIBMAG
-#define RANLIBMAG	"__.SYMDEF"
-#endif
+#include    "dir.h"
+#include    "config.h"
 
 static Lst	  archives;   /* Lst of archives we've already examined */
 
@@ -108,9 +120,54 @@ typedef struct Arch {
     char	  *name;      /* Name of archive */
     Hash_Table	  members;    /* All the members of the archive described
 			       * by <name, struct ar_hdr *> key/value pairs */
+    char	  *fnametab;  /* Extended name table strings */
+    size_t	  fnamesize;  /* Size of the string table */
 } Arch;
 
-static FILE *ArchFindMember();
+static int ArchFindArchive __P((ClientData, ClientData));
+static void ArchFree __P((ClientData));
+static struct ar_hdr *ArchStatMember __P((char *, char *, Boolean));
+static FILE *ArchFindMember __P((char *, char *, struct ar_hdr *, char *));
+#if defined(__svr4__) || defined(__SVR4) || defined(__ELF__)
+#define SVR4ARCHIVES
+static int ArchSVR4Entry __P((Arch *, char *, size_t, FILE *));
+#endif
+
+/*-
+ *-----------------------------------------------------------------------
+ * ArchFree --
+ *	Free memory used by an archive
+ *
+ * Results:
+ *	None.
+ *
+ * Side Effects:
+ *	None.
+ *
+ *-----------------------------------------------------------------------
+ */
+static void
+ArchFree(ap)
+    ClientData ap;
+{
+    Arch *a = (Arch *) ap;
+    Hash_Search	  search;
+    Hash_Entry	  *entry;
+
+    /* Free memory from hash entries */
+    for (entry = Hash_EnumFirst(&a->members, &search);
+	 entry != (Hash_Entry *)NULL;
+	 entry = Hash_EnumNext(&search))
+	free((Address) Hash_GetValue (entry));
+
+    free(a->name);
+    if (a->fnametab)
+	free(a->fnametab);
+    Hash_DeleteTable(&a->members);
+    free((Address) a);
+}
+
+
 
 /*-
  *-----------------------------------------------------------------------
@@ -139,13 +196,13 @@ Arch_ParseArchive (linePtr, nodeLst, ctxt)
     GNode	    *gn;     	    /* New node */
     char	    *libName;  	    /* Library-part of specification */
     char	    *memName;  	    /* Member-part of specification */
-    char	    nameBuf[BSIZE]; /* temporary place for node name */
+    char	    nameBuf[MAKE_BSIZE]; /* temporary place for node name */
     char	    saveChar;  	    /* Ending delimiter of member-name */
     Boolean 	    subLibName;	    /* TRUE if libName should have/had
 				     * variable substitution performed on it */
 
     libName = *linePtr;
-    
+
     subLibName = FALSE;
 
     for (cp = libName; *cp != '(' && *cp != '\0'; cp++) {
@@ -157,14 +214,14 @@ Arch_ParseArchive (linePtr, nodeLst, ctxt)
 	    int 	length;
 	    Boolean	freeIt;
 	    char	*result;
-	    
+
 	    result=Var_Parse(cp, ctxt, TRUE, &length, &freeIt);
 	    if (result == var_Error) {
 		return(FAILURE);
 	    } else {
 		subLibName = TRUE;
 	    }
-	    
+
 	    if (freeIt) {
 		free(result);
 	    }
@@ -174,11 +231,11 @@ Arch_ParseArchive (linePtr, nodeLst, ctxt)
 
     *cp++ = '\0';
     if (subLibName) {
-	libName = Var_Subst(libName, ctxt, TRUE);
+	libName = Var_Subst(NULL, libName, ctxt, TRUE);
     }
-    
 
-    while (1) {
+
+    for (;;) {
 	/*
 	 * First skip to the start of the member's name, mark that
 	 * place and skip to the end of it (either white-space or
@@ -252,8 +309,8 @@ Arch_ParseArchive (linePtr, nodeLst, ctxt)
 	    char    *buf;
 	    char    *sacrifice;
 	    char    *oldMemName = memName;
-	    
-	    memName = Var_Subst(memName, ctxt, TRUE);
+
+	    memName = Var_Subst(NULL, memName, ctxt, TRUE);
 
 	    /*
 	     * Now form an archive spec and recurse to deal with nested
@@ -264,7 +321,7 @@ Arch_ParseArchive (linePtr, nodeLst, ctxt)
 
 	    sprintf(buf, "%s(%s)", libName, memName);
 
-	    if (index(memName, '$') && strcmp(memName, oldMemName) == 0) {
+	    if (strchr(memName, '$') && strcmp(memName, oldMemName) == 0) {
 		/*
 		 * Must contain dynamic sources, so we can't deal with it now.
 		 * Just create an ARCHV node for the thing and let
@@ -298,7 +355,7 @@ Arch_ParseArchive (linePtr, nodeLst, ctxt)
 	    Dir_Expand(memName, dirSearchPath, members);
 	    while (!Lst_IsEmpty(members)) {
 		member = (char *)Lst_DeQueue(members);
-		
+
 		sprintf(nameBuf, "%s(%s)", libName, member);
 		free(member);
 		gn = Targ_FindNode (nameBuf, TARG_CREATE);
@@ -337,7 +394,7 @@ Arch_ParseArchive (linePtr, nodeLst, ctxt)
 	if (doSubst) {
 	    free(memName);
 	}
-	
+
 	*cp = saveChar;
     }
 
@@ -377,10 +434,10 @@ Arch_ParseArchive (linePtr, nodeLst, ctxt)
  */
 static int
 ArchFindArchive (ar, archName)
-    Arch	  *ar;	      	  /* Current list element */
-    char	  *archName;  	  /* Name we want */
+    ClientData	  ar;	      	  /* Current list element */
+    ClientData	  archName;  	  /* Name we want */
 {
-    return (strcmp (archName, ar->name));
+    return (strcmp ((char *) archName, ((Arch *) ar)->name));
 }
 
 /*-
@@ -413,19 +470,12 @@ ArchStatMember (archive, member, hash)
     int		  size;       /* Size of archive member */
     char	  *cp;	      /* Useful character pointer */
     char	  magic[SARMAG];
-    int		  len;
     LstNode	  ln;	      /* Lst member containing archive descriptor */
     Arch	  *ar;	      /* Archive descriptor */
     Hash_Entry	  *he;	      /* Entry containing member's description */
     struct ar_hdr arh;        /* archive-member header for reading archive */
-    char	  memName[AR_MAX_NAME_LEN+1];
-    	    	    	    /* Current member name while hashing. The name is
-			     * truncated to AR_MAX_NAME_LEN bytes, but we need
-			     * room for the null byte... */
-    char    	  copy[AR_MAX_NAME_LEN+1];
-    	    	    	    /* Holds copy of last path element from member, if
-			     * it has to be truncated, so we don't have to
-			     * figure it out again once the table is hashed. */
+    char	  memName[MAXPATHLEN+1];
+    	    	    	    /* Current member name while hashing. */
 
     /*
      * Because of space constraints and similar things, files are archived
@@ -433,17 +483,10 @@ ArchStatMember (archive, member, hash)
      * to point 'member' to the final component, if there is one, to make
      * the comparisons easier...
      */
-    cp = rindex (member, '/');
+    cp = strrchr (member, '/');
     if (cp != (char *) NULL) {
 	member = cp + 1;
     }
-    len = strlen (member);
-    if (len > AR_MAX_NAME_LEN) {
-	len = AR_MAX_NAME_LEN;
-	strncpy(copy, member, AR_MAX_NAME_LEN);
-	copy[AR_MAX_NAME_LEN] = '\0';
-	member = copy;
-    }
 
     ln = Lst_Find (archives, (ClientData) archive, ArchFindArchive);
     if (ln != NILLNODE) {
@@ -454,6 +497,17 @@ ArchStatMember (archive, member, hash)
 	if (he != (Hash_Entry *) NULL) {
 	    return ((struct ar_hdr *) Hash_GetValue (he));
 	} else {
+	    /* Try truncated name */
+	    char copy[AR_MAX_NAME_LEN+1];
+	    int len = strlen (member);
+
+	    if (len > AR_MAX_NAME_LEN) {
+		len = AR_MAX_NAME_LEN;
+		strncpy(copy, member, AR_MAX_NAME_LEN);
+		copy[AR_MAX_NAME_LEN] = '\0';
+	    }
+	    if ((he = Hash_FindEntry (&ar->members, copy)) != NULL)
+		return ((struct ar_hdr *) Hash_GetValue (he));
 	    return ((struct ar_hdr *) NULL);
 	}
     }
@@ -486,7 +540,7 @@ ArchStatMember (archive, member, hash)
     if (arch == (FILE *) NULL) {
 	return ((struct ar_hdr *) NULL);
     }
-    
+
     /*
      * We use the ARMAG string to make sure this is an archive we
      * can handle...
@@ -498,42 +552,86 @@ ArchStatMember (archive, member, hash)
     }
 
     ar = (Arch *)emalloc (sizeof (Arch));
-    ar->name = strdup (archive);
+    ar->name = estrdup (archive);
+    ar->fnametab = NULL;
+    ar->fnamesize = 0;
     Hash_InitTable (&ar->members, -1);
     memName[AR_MAX_NAME_LEN] = '\0';
-    
+
     while (fread ((char *)&arh, sizeof (struct ar_hdr), 1, arch) == 1) {
 	if (strncmp ( arh.ar_fmag, ARFMAG, sizeof (arh.ar_fmag)) != 0) {
-				 /*
-				  * The header is bogus, so the archive is bad
-				  * and there's no way we can recover...
-				  */
-				 fclose (arch);
-				 Hash_DeleteTable (&ar->members);
-				 free ((Address)ar);
-				 return ((struct ar_hdr *) NULL);
+	    /*
+	     * The header is bogus, so the archive is bad
+	     * and there's no way we can recover...
+	     */
+	    goto badarch;
 	} else {
+	    /*
+	     * We need to advance the stream's pointer to the start of the
+	     * next header. Files are padded with newlines to an even-byte
+	     * boundary, so we need to extract the size of the file from the
+	     * 'size' field of the header and round it up during the seek.
+	     */
+	    arh.ar_size[sizeof(arh.ar_size)-1] = '\0';
+	    size = (int) strtol(arh.ar_size, NULL, 10);
+
 	    (void) strncpy (memName, arh.ar_name, sizeof(arh.ar_name));
 	    for (cp = &memName[AR_MAX_NAME_LEN]; *cp == ' '; cp--) {
 		continue;
 	    }
 	    cp[1] = '\0';
 
-	    he = Hash_CreateEntry (&ar->members, strdup (memName),
-				   (Boolean *)NULL);
+#ifdef SVR4ARCHIVES
+	    /*
+	     * svr4 names are slash terminated. Also svr4 extended AR format.
+	     */
+	    if (memName[0] == '/') {
+		/*
+		 * svr4 magic mode; handle it
+		 */
+		switch (ArchSVR4Entry(ar, memName, size, arch)) {
+		case -1:  /* Invalid data */
+		    goto badarch;
+		case 0:	  /* List of files entry */
+		    continue;
+		default:  /* Got the entry */
+		    break;
+		}
+	    }
+	    else {
+		if (cp[0] == '/')
+		    cp[0] = '\0';
+	    }
+#endif
+
+#ifdef AR_EFMT1
+	    /*
+	     * BSD 4.4 extended AR format: #1/<namelen>, with name as the
+	     * first <namelen> bytes of the file
+	     */
+	    if (strncmp(memName, AR_EFMT1, sizeof(AR_EFMT1) - 1) == 0 &&
+		isdigit(memName[sizeof(AR_EFMT1) - 1])) {
+
+		unsigned int elen = atoi(&memName[sizeof(AR_EFMT1)-1]);
+
+		if (elen > MAXPATHLEN)
+			goto badarch;
+		if (fread (memName, elen, 1, arch) != 1)
+			goto badarch;
+		memName[elen] = '\0';
+		fseek (arch, -elen, SEEK_CUR);
+		if (DEBUG(ARCH) || DEBUG(MAKE)) {
+		    printf("ArchStat: Extended format entry for %s\n", memName);
+		}
+	    }
+#endif
+
+	    he = Hash_CreateEntry (&ar->members, memName, (Boolean *)NULL);
 	    Hash_SetValue (he, (ClientData)emalloc (sizeof (struct ar_hdr)));
-	    bcopy ((Address)&arh, (Address)Hash_GetValue (he), 
+	    memcpy ((Address)Hash_GetValue (he), (Address)&arh,
 		sizeof (struct ar_hdr));
 	}
-	/*
-	 * We need to advance the stream's pointer to the start of the
-	 * next header. Files are padded with newlines to an even-byte
-	 * boundary, so we need to extract the size of the file from the
-	 * 'size' field of the header and round it up during the seek.
-	 */
-	arh.ar_size[sizeof(arh.ar_size)-1] = '\0';
-	(void) sscanf (arh.ar_size, "%10d", &size);
-	fseek (arch, (size + 1) & ~1, 1);
+	fseek (arch, (size + 1) & ~1, SEEK_CUR);
     }
 
     fclose (arch);
@@ -551,8 +649,122 @@ ArchStatMember (archive, member, hash)
     } else {
 	return ((struct ar_hdr *) NULL);
     }
+
+badarch:
+    fclose (arch);
+    Hash_DeleteTable (&ar->members);
+    if (ar->fnametab)
+	free(ar->fnametab);
+    free ((Address)ar);
+    return ((struct ar_hdr *) NULL);
 }
 
+#ifdef SVR4ARCHIVES
+/*-
+ *-----------------------------------------------------------------------
+ * ArchSVR4Entry --
+ *	Parse an SVR4 style entry that begins with a slash.
+ *	If it is "//", then load the table of filenames
+ *	If it is "/<offset>", then try to substitute the long file name
+ *	from offset of a table previously read.
+ *
+ * Results:
+ *	-1: Bad data in archive
+ *	 0: A table was loaded from the file
+ *	 1: Name was successfully substituted from table
+ *	 2: Name was not successfully substituted from table
+ *
+ * Side Effects:
+ *	If a table is read, the file pointer is moved to the next archive
+ *	member
+ *
+ *-----------------------------------------------------------------------
+ */
+static int
+ArchSVR4Entry(ar, name, size, arch)
+	Arch *ar;
+	char *name;
+	size_t size;
+	FILE *arch;
+{
+#define ARLONGNAMES1 "//"
+#define ARLONGNAMES2 "/ARFILENAMES"
+    size_t entry;
+    char *ptr, *eptr;
+
+    if (strncmp(name, ARLONGNAMES1, sizeof(ARLONGNAMES1) - 1) == 0 ||
+	strncmp(name, ARLONGNAMES2, sizeof(ARLONGNAMES2) - 1) == 0) {
+
+	if (ar->fnametab != NULL) {
+	    if (DEBUG(ARCH)) {
+		printf("Attempted to redefine an SVR4 name table\n");
+	    }
+	    return -1;
+	}
+
+	/*
+	 * This is a table of archive names, so we build one for
+	 * ourselves
+	 */
+	ar->fnametab = emalloc(size);
+	ar->fnamesize = size;
+
+	if (fread(ar->fnametab, size, 1, arch) != 1) {
+	    if (DEBUG(ARCH)) {
+		printf("Reading an SVR4 name table failed\n");
+	    }
+	    return -1;
+	}
+	eptr = ar->fnametab + size;
+	for (entry = 0, ptr = ar->fnametab; ptr < eptr; ptr++)
+	    switch (*ptr) {
+	    case '/':
+		entry++;
+		*ptr = '\0';
+		break;
+
+	    case '\n':
+		break;
+
+	    default:
+		break;
+	    }
+	if (DEBUG(ARCH)) {
+	    printf("Found svr4 archive name table with %lu entries\n",
+	            (u_long)entry);
+	}
+	return 0;
+    }
+
+    if (name[1] == ' ' || name[1] == '\0')
+	return 2;
+
+    entry = (size_t) strtol(&name[1], &eptr, 0);
+    if ((*eptr != ' ' && *eptr != '\0') || eptr == &name[1]) {
+	if (DEBUG(ARCH)) {
+	    printf("Could not parse SVR4 name %s\n", name);
+	}
+	return 2;
+    }
+    if (entry >= ar->fnamesize) {
+	if (DEBUG(ARCH)) {
+	    printf("SVR4 entry offset %s is greater than %lu\n",
+		   name, (u_long)ar->fnamesize);
+	}
+	return 2;
+    }
+
+    if (DEBUG(ARCH)) {
+	printf("Replaced %s with %s\n", name, &ar->fnametab[entry]);
+    }
+
+    (void) strncpy(name, &ar->fnametab[entry], MAXPATHLEN);
+    name[MAXPATHLEN] = '\0';
+    return 1;
+}
+#endif
+
+
 /*-
  *-----------------------------------------------------------------------
  * ArchFindMember --
@@ -582,13 +794,13 @@ ArchFindMember (archive, member, arhPtr, mode)
     int		  size;       /* Size of archive member */
     char	  *cp;	      /* Useful character pointer */
     char	  magic[SARMAG];
-    int		  len;
+    int		  len, tlen;
 
     arch = fopen (archive, mode);
     if (arch == (FILE *) NULL) {
 	return ((FILE *) NULL);
     }
-    
+
     /*
      * We use the ARMAG string to make sure this is an archive we
      * can handle...
@@ -605,15 +817,15 @@ ArchFindMember (archive, member, arhPtr, mode)
      * to point 'member' to the final component, if there is one, to make
      * the comparisons easier...
      */
-    cp = rindex (member, '/');
+    cp = strrchr (member, '/');
     if (cp != (char *) NULL) {
 	member = cp + 1;
     }
-    len = strlen (member);
+    len = tlen = strlen (member);
     if (len > sizeof (arhPtr->ar_name)) {
-	len = sizeof (arhPtr->ar_name);
+	tlen = sizeof (arhPtr->ar_name);
     }
-    
+
     while (fread ((char *)arhPtr, sizeof (struct ar_hdr), 1, arch) == 1) {
 	if (strncmp(arhPtr->ar_fmag, ARFMAG, sizeof (arhPtr->ar_fmag) ) != 0) {
 	     /*
@@ -622,7 +834,7 @@ ArchFindMember (archive, member, arhPtr, mode)
 	      */
 	     fclose (arch);
 	     return ((FILE *) NULL);
-	} else if (strncmp (member, arhPtr->ar_name, len) == 0) {
+	} else if (strncmp (member, arhPtr->ar_name, tlen) == 0) {
 	    /*
 	     * If the member's name doesn't take up the entire 'name' field,
 	     * we have to be careful of matching prefixes. Names are space-
@@ -630,8 +842,8 @@ ArchFindMember (archive, member, arhPtr, mode)
 	     * of the matched string is anything but a space, this isn't the
 	     * member we sought.
 	     */
-	    if (len != sizeof(arhPtr->ar_name) && arhPtr->ar_name[len] != ' '){
-		continue;
+	    if (tlen != sizeof(arhPtr->ar_name) && arhPtr->ar_name[tlen] != ' '){
+		goto skip;
 	    } else {
 		/*
 		 * To make life easier, we reposition the file at the start
@@ -640,10 +852,45 @@ ArchFindMember (archive, member, arhPtr, mode)
 		 * the file at the actual member, rather than its header, but
 		 * not here...
 		 */
-		fseek (arch, -sizeof(struct ar_hdr), 1);
+		fseek (arch, -sizeof(struct ar_hdr), SEEK_CUR);
 		return (arch);
 	    }
-	} else {
+	} else
+#ifdef AR_EFMT1
+		/*
+		 * BSD 4.4 extended AR format: #1/<namelen>, with name as the
+		 * first <namelen> bytes of the file
+		 */
+	    if (strncmp(arhPtr->ar_name, AR_EFMT1,
+					sizeof(AR_EFMT1) - 1) == 0 &&
+		isdigit(arhPtr->ar_name[sizeof(AR_EFMT1) - 1])) {
+
+		unsigned int elen = atoi(&arhPtr->ar_name[sizeof(AR_EFMT1)-1]);
+		char ename[MAXPATHLEN];
+
+		if (elen > MAXPATHLEN) {
+			fclose (arch);
+			return NULL;
+		}
+		if (fread (ename, elen, 1, arch) != 1) {
+			fclose (arch);
+			return NULL;
+		}
+		ename[elen] = '\0';
+		if (DEBUG(ARCH) || DEBUG(MAKE)) {
+		    printf("ArchFind: Extended format entry for %s\n", ename);
+		}
+		if (strncmp(ename, member, len) == 0) {
+			/* Found as extended name */
+			fseek (arch, -sizeof(struct ar_hdr) - elen, SEEK_CUR);
+			return (arch);
+		}
+		fseek (arch, -elen, SEEK_CUR);
+		goto skip;
+	} else
+#endif
+	{
+skip:
 	    /*
 	     * This isn't the member we're after, so we need to advance the
 	     * stream's pointer to the start of the next header. Files are
@@ -652,8 +899,8 @@ ArchFindMember (archive, member, arhPtr, mode)
 	     * header and round it up during the seek.
 	     */
 	    arhPtr->ar_size[sizeof(arhPtr->ar_size)-1] = '\0';
-	    (void)sscanf (arhPtr->ar_size, "%10d", &size);
-	    fseek (arch, (size + 1) & ~1, 1);
+	    size = (int) strtol(arhPtr->ar_size, NULL, 10);
+	    fseek (arch, (size + 1) & ~1, SEEK_CUR);
 	}
     }
 
@@ -686,11 +933,16 @@ Arch_Touch (gn)
 {
     FILE *	  arch;	  /* Stream open to archive, positioned properly */
     struct ar_hdr arh;	  /* Current header describing member */
+    char *p1, *p2;
 
-    arch = ArchFindMember(Var_Value (ARCHIVE, gn),
-			  Var_Value (TARGET, gn),
+    arch = ArchFindMember(Var_Value (ARCHIVE, gn, &p1),
+			  Var_Value (MEMBER, gn, &p2),
 			  &arh, "r+");
-    sprintf(arh.ar_date, "%-12d", now);
+    if (p1)
+	free(p1);
+    if (p2)
+	free(p2);
+    sprintf(arh.ar_date, "%-12ld", (long) now);
 
     if (arch != (FILE *) NULL) {
 	(void)fwrite ((char *)&arh, sizeof (struct ar_hdr), 1, arch);
@@ -717,21 +969,22 @@ void
 Arch_TouchLib (gn)
     GNode	    *gn;      	/* The node of the library to touch */
 {
+#ifdef RANLIBMAG
     FILE *	    arch;	/* Stream open to archive */
     struct ar_hdr   arh;      	/* Header describing table of contents */
-    struct timeval  times[2];	/* Times for utimes() call */
+    struct utimbuf  times;	/* Times for utime() call */
 
     arch = ArchFindMember (gn->path, RANLIBMAG, &arh, "r+");
-    sprintf(arh.ar_date, "%-12d", now);
+    sprintf(arh.ar_date, "%-12ld", (long) now);
 
     if (arch != (FILE *) NULL) {
 	(void)fwrite ((char *)&arh, sizeof (struct ar_hdr), 1, arch);
 	fclose (arch);
 
-	times[0].tv_sec = times[1].tv_sec = now;
-	times[0].tv_usec = times[1].tv_usec = 0;
-	utimes(gn->path, times);
+	times.actime = times.modtime = now;
+	utime(gn->path, &times);
     }
+#endif
 }
 
 /*-
@@ -748,18 +1001,24 @@ Arch_TouchLib (gn)
  *
  *-----------------------------------------------------------------------
  */
-int
+time_t
 Arch_MTime (gn)
     GNode	  *gn;	      /* Node describing archive member */
 {
     struct ar_hdr *arhPtr;    /* Header of desired member */
-    int		  modTime;    /* Modification time as an integer */
+    time_t	  modTime;    /* Modification time as an integer */
+    char *p1, *p2;
 
-    arhPtr = ArchStatMember (Var_Value (ARCHIVE, gn),
-			     Var_Value (TARGET, gn),
+    arhPtr = ArchStatMember (Var_Value (ARCHIVE, gn, &p1),
+			     Var_Value (MEMBER, gn, &p2),
 			     TRUE);
+    if (p1)
+	free(p1);
+    if (p2)
+	free(p2);
+
     if (arhPtr != (struct ar_hdr *) NULL) {
-	(void)sscanf (arhPtr->ar_date, "%12d", &modTime);
+	modTime = (time_t) strtol(arhPtr->ar_date, NULL, 10);
     } else {
 	modTime = 0;
     }
@@ -782,7 +1041,7 @@ Arch_MTime (gn)
  *
  *-----------------------------------------------------------------------
  */
-int
+time_t
 Arch_MemMTime (gn)
     GNode   	  *gn;
 {
@@ -806,8 +1065,8 @@ Arch_MemMTime (gn)
 	     * child. We keep searching its parents in case some other
 	     * parent requires this child to exist...
 	     */
-	    nameStart = index (pgn->name, '(') + 1;
-	    nameEnd = index (nameStart, ')');
+	    nameStart = strchr (pgn->name, '(') + 1;
+	    nameEnd = strchr (nameStart, ')');
 
 	    if (pgn->make &&
 		strncmp(nameStart, gn->name, nameEnd - nameStart) == 0) {
@@ -831,7 +1090,7 @@ Arch_MemMTime (gn)
 /*-
  *-----------------------------------------------------------------------
  * Arch_FindLib --
- *	Search for a library along the given search path. 
+ *	Search for a library along the given search path.
  *
  * Results:
  *	None.
@@ -866,7 +1125,7 @@ Arch_FindLib (gn, path)
     Var_Set (TARGET, gn->name, gn);
 #else
     Var_Set (TARGET, gn->path == (char *) NULL ? gn->name : gn->path, gn);
-#endif LIBRARIES
+#endif /* LIBRARIES */
 }
 
 /*-
@@ -888,10 +1147,14 @@ Arch_FindLib (gn, path)
  *	    Its modification time is greater than the time at which the
  *	    	  make began (i.e. it's been modified in the course
  *	    	  of the make, probably by archiving).
- *	    Its modification time doesn't agree with the modification
- *	    	  time of its RANLIBMAG member (i.e. its table of contents
- *	    	  is out-of-date).
- *
+ *	    The modification time of one of its sources is greater than
+ *		  the one of its RANLIBMAG member (i.e. its table of contents
+ *	    	  is out-of-date). We don't compare of the archive time
+ *		  vs. TOC time because they can be too close. In my
+ *		  opinion we should not bother with the TOC at all since
+ *		  this is used by 'ar' rules that affect the data contents
+ *		  of the archive, not by ranlib rules, which affect the
+ *		  TOC.
  *
  * Results:
  *	TRUE if the library is out-of-date. FALSE otherwise.
@@ -906,24 +1169,27 @@ Arch_LibOODate (gn)
     GNode   	  *gn;  	/* The library's graph node */
 {
     Boolean 	  oodate;
-    
-    if (OP_NOP(gn->type) && Lst_IsEmpty(gn->children)) {
+
+    if (gn->type & OP_PHONY) {
+	oodate = TRUE;
+    } else if (OP_NOP(gn->type) && Lst_IsEmpty(gn->children)) {
 	oodate = FALSE;
     } else if ((gn->mtime > now) || (gn->mtime < gn->cmtime)) {
 	oodate = TRUE;
     } else {
+#ifdef RANLIBMAG
 	struct ar_hdr  	*arhPtr;    /* Header for __.SYMDEF */
 	int 	  	modTimeTOC; /* The table-of-contents's mod time */
 
 	arhPtr = ArchStatMember (gn->path, RANLIBMAG, FALSE);
 
 	if (arhPtr != (struct ar_hdr *)NULL) {
-	    (void)sscanf (arhPtr->ar_date, "%12d", &modTimeTOC);
+	    modTimeTOC = (int) strtol(arhPtr->ar_date, NULL, 10);
 
 	    if (DEBUG(ARCH) || DEBUG(MAKE)) {
 		printf("%s modified %s...", RANLIBMAG, Targ_FmtTime(modTimeTOC));
 	    }
-	    oodate = (gn->mtime > modTimeTOC);
+	    oodate = (gn->cmtime > modTimeTOC);
 	} else {
 	    /*
 	     * A library w/o a table of contents is out-of-date
@@ -933,6 +1199,9 @@ Arch_LibOODate (gn)
 	    }
 	    oodate = TRUE;
 	}
+#else
+	oodate = FALSE;
+#endif
     }
     return (oodate);
 }
@@ -955,3 +1224,58 @@ Arch_Init ()
 {
     archives = Lst_Init (FALSE);
 }
+
+
+
+/*-
+ *-----------------------------------------------------------------------
+ * Arch_End --
+ *	Cleanup things for this module.
+ *
+ * Results:
+ *	None.
+ *
+ * Side Effects:
+ *	The 'archives' list is freed
+ *
+ *-----------------------------------------------------------------------
+ */
+void
+Arch_End ()
+{
+    Lst_Destroy(archives, ArchFree);
+}
+
+/*-
+ *-----------------------------------------------------------------------
+ * Arch_IsLib --
+ *	Check if the node is a library
+ *
+ * Results:
+ *	True or False.
+ *
+ * Side Effects:
+ *	None.
+ *
+ *-----------------------------------------------------------------------
+ */
+int
+Arch_IsLib(gn)
+    GNode *gn;
+{
+    static const char armag[] = "!<arch>\n";
+    char buf[sizeof(armag)-1];
+    int fd;
+
+    if ((fd = open(gn->path, O_RDONLY)) == -1)
+	return FALSE;
+
+    if (read(fd, buf, sizeof(buf)) != sizeof(buf)) {
+	(void) close(fd);
+	return FALSE;
+    }
+
+    (void) close(fd);
+
+    return memcmp(buf, armag, sizeof(buf)) == 0;
+}
